#!/usr/bin/env python3

"""Stuff that doesn't go anywhere else
"""

import asyncio
import logging
import os
import time
from collections.abc import AsyncIterator, Callable, Coroutine, Iterator
from functools import wraps
from pathlib import Path
from typing import NoReturn, cast

from asyncinotify import Event, Inotify, Mask


def log() -> logging.Logger:
    """Logger for this module"""
    return logging.getLogger("medit.misc")


def watchdog(
    afunc: Callable[..., Coroutine[object, object, object]]
) -> Callable[..., Coroutine[object, object, object]]:
    """Watch for async functions to throw an unhandled exception"""

    @wraps(afunc)
    async def run(*args: object, **kwargs: object) -> object:
        """Run wrapped function and handle exceptions"""
        try:
            return await afunc(*args, **kwargs)
        except asyncio.CancelledError:
            log().info("Task cancelled: `%s`", afunc.__name__)
        except KeyboardInterrupt:
            log().info("KeyboardInterrupt in `%s`", afunc.__name__)
        except Exception:  # pylint: disable=broad-except
            log().exception("Exception in `%s`:", afunc.__name__)
            asyncio.get_event_loop().stop()
        return None

    return run


def impatient(func):
    @wraps(func)
    def run(*args: object, **kwargs: object) -> object:
        try:
            t1 = time.time()
            return func(*args, **kwargs)
        finally:
            if (duration := time.time() - t1) > 0.1:
                log().warn("%s took %.2fs!", func.__name__, duration)

    return run


async def fs_changes(
    *paths: Path,
    queue: asyncio.Queue[str] = asyncio.Queue(),
    mask: Mask = Mask.CLOSE_WRITE
    | Mask.MOVED_TO
    | Mask.CREATE
    | Mask.MODIFY
    | Mask.MOVE
    | Mask.DELETE
    | Mask.MOVE_SELF,
    postpone: bool = False,
    timeout: float = 2,
) -> AsyncIterator[Path]:
    """Controllable, timed filesystem watcher"""

    # pylint: disable=too-many-locals

    async def fuse_fn(queue: asyncio.Queue[str], timeout: float) -> None:
        await asyncio.sleep(timeout)
        await queue.put("timeout")

    def expand_paths(path: Path, recursive: bool = True) -> Iterator[Path]:
        yield path
        if path.is_dir() and recursive:
            for file_or_directory in path.rglob("*"):
                if file_or_directory.is_dir():
                    yield file_or_directory

    def task(name: str) -> asyncio.Task[str | Event]:
        """Creates a task from a name identifying a data source to read from"""
        return asyncio.create_task(
            cast(asyncio.Queue[str] | Inotify, {"inotify": inotify, "mqueue": queue}[name]).get(),
            name=name,
        )

    with Inotify() as inotify:
        for path in set(sub_path.absolute() for p in paths for sub_path in expand_paths(Path(p))):
            log().debug("add fs watch for %s", path)
            inotify.add_watch(path, mask)
        fuse = None
        changed_files = set()
        tasks = set(map(task, ("inotify", "mqueue")))

        while True:
            done, tasks = await asyncio.wait(
                fs=tasks,
                return_when=asyncio.FIRST_COMPLETED,
            )
            for event in done:
                event_type, event_value = event.get_name(), event.result()
                tasks.add(task(event_type))
                if event_type == "inotify":
                    assert isinstance(event_value, Event)
                    if event_value.path:
                        changed_files.add(event_value.path)
                    if postpone and fuse:
                        fuse.cancel()
                        del fuse
                        fuse = None
                    if not fuse:
                        fuse = asyncio.create_task(fuse_fn(queue, timeout))
                elif event_type == "mqueue":
                    if event_value == "timeout":
                        del fuse
                        fuse = None
                        for file in changed_files:
                            yield file
                        changed_files.clear()


def setup_logging(level: str | int = logging.DEBUG) -> None:
    '''
    def thread_id_filter(record):
        """Inject thread_id to log records"""
        record.thread_id = threading.get_native_id()
        return record

    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter("%(asctime)s | %(levelname)s | %(thread_id)s | %(message)s")
    )
    handler.addFilter(thread_id_filter)
    logger().addHandler(handler)
    logging.getLogger().setLevel(level)
    '''
    use_col = "TERM" in os.environ
    col_terminator = "\033[0m" if use_col else ""
    logging.basicConfig(
        format=f"%(levelname)s %(asctime)s.%(msecs)03d %(name)-12s│ %(message)s{col_terminator}",
        datefmt="%H:%M:%S",
        level=getattr(logging, level) if isinstance(level, str) else level,
    )
    for name, color in (
        ("DEBUG", "\033[32m"),
        ("INFO", "\033[36m"),
        ("WARNING", "\033[33m"),
        ("ERROR", "\033[31m"),
        ("CRITICAL", "\033[37m"),
    ):
        logging.addLevelName(
            getattr(logging, name),
            f"{color if use_col else ''}({name[0] * 2})",
        )


def throw(exc: Exception) -> NoReturn:
    """Make raising an exception functional"""
    raise exc



#!/usr/bin/env python3

from pathlib import Path

import pygments.lexers
from markdown import markdown
from markdown_include.include import MarkdownInclude
from pygments import highlight
from pygments.formatters import HtmlFormatter
from PyQt6.QtWebEngineWidgets import QWebEngineView


class MView(QWebEngineView):
    def __init__(self, parent):
        super().__init__(parent)
        self.css_content_markdown = "".join(
            f"<style>{open(Path(__file__).parent / path).read()}</style>"
            for path in (
                "styles/external/github-markdown-light.css",
                "styles/external/thomasf-solarized-css/solarized-light.css",
                "styles/external/solarized-light-all-sites.css",
                "styles/external/pygments-solarized-style/solarized-light.css",
                # "styles/external/github-markdown-dark.css",
                # "styles/external/thomasf-solarized-css/solarized-dark.css",
                # "styles/external/solarized-dark-all-sites.css",
                # "styles/external/pygments-solarized-style/solarized-dark.css",
            )
        )

        self.css_content = "".join(
            f"<style>{open(Path(__file__).parent / path).read()}</style>"
            for path in ("styles/external/pygments-solarized-style/solarized-light.css",)
        )

        self.css_content_unused = "".join(
            f"<style>{open(Path(__file__).parent / path).read()}</style>"
            for path in (
                #
                # Basic markdown stuff
                #
                # GitHub - html only
                # "styles/external/github-markdown-light.css",
                # "styles/external/github-markdown-dark.css",
                # dark - https://nihaojob.github.io/markdown-css-smartisan/
                # "styles/external/github-markdown.css",
                # html only, fonts not affected `.markdown` class
                # https://cdn.jsdelivr.net/npm/@naokim03/markdown-theme-solarized@1.0.2/
                # "styles/markdown.min.css",
                # https://github.com/alphapapa/solarized-everything-css
                # "styles/external/solarized-light-all-sites.css",
                # "styles/external/solarized-dark-all-sites.css",
                # html only
                # https://github.com/thomasf/solarized-css
                # "styles/solarized-light.css",
                # "styles/solarized-dark.css",  #               | light, transparent, decent
                #
                # Pygments
                #
                # pygments via `pygments-solarized-style`
                # "styles/external/solarized-light.css",  #               | light, transparent, decent
                # "styles/external/solarized-dark.css",  #               | light, transparent, decent
                # pygments defaults
                # "styles/pygments/abap.css",  #               | light, transparent, decent
                # "styles/pygments/algol.css",  #              | light/dark, transparent, bw
                # "styles/pygments/algol_nu.css",  #           | light/dark, transparent, bw
                # "styles/pygments/arduino.css",  #            | light/dark, decent
                # "styles/pygments/autumn.css",  #             | light (dark)
                # "styles/pygments/borland.css",  #            | light
                # "styles/pygments/bw.css",  #                 |
                # "styles/pygments/colorful.css",  #           | light +string
                # "styles/pygments/default.css",  #            |
                # "styles/pygments/dracula.css",  #            | dark
                # "styles/pygments/emacs.css",  #              |
                # "styles/pygments/friendly.css",  #           |
                # "styles/pygments/friendly_grayscale.css",  # |
                # "styles/pygments/fruity.css",  #             | dark
                # "styles/pygments/igor.css",  #               | light (dark)
                # "styles/pygments/inkpot.css",  #             | dark (light) +strings
                # "styles/pygments/lilypond.css",  #           | light dark
                # "styles/pygments/lovelace.css",  #           | light dark
                # "styles/pygments/manni.css",  #              | dark light +nice
                # "styles/pygments/material.css",  #           | dark
                # "styles/pygments/monokai.css",  #            | dark
                # "styles/pygments/murphy.css",  #             |
                # "styles/pygments/native.css",  #             |
                # "styles/pygments/nord.css",  #               |
                # "styles/pygments/pastie.css",  #             |
                # "styles/pygments/perldoc.css",  #            |
                # "styles/pygments/rainbow_dash.css",  #       |
                # "styles/pygments/rrt.css",  #                |
                # "styles/pygments/sas.css",  #                |
                # "styles/pygments/staroffice.css",  #         |
                # "styles/pygments/tango.css",  #              |
                # "styles/pygments/trac.css",  #               | light dark
                # "styles/pygments/vim.css",  #                |
                # "styles/pygments/vs.css",  #                 | light, transparent, decent
                # "styles/pygments/xcode.css",  #              | light, transparent, decent
                # "styles/pygments/zenburn.css",  #            | dark
            )
        )

    def show(self, content, highlight_mode, base_dir: Path):
        if highlight_mode == ".md":
            html, style = (
                markdown(
                    content,
                    extensions=[
                        "extra",
                        "codehilite",
                        "markdown_checklist.extension",
                        #"md_mermaid",
                        MarkdownInclude(
                            configs={
                                "base_path": base_dir,
                                "encoding": "utf-8",
                            }
                        ),
                    ],
                ),
                self.css_content_markdown,
            )
        else:
            formatter = HtmlFormatter(linenos="table", cssclass="codehilite")
            lexer = (
                pygments.lexers.YamlLexer()
                if highlight_mode in {".yaml", ".yml"}
                else pygments.lexers.BashLexer()
                if highlight_mode in {".sh"}
                else pygments.lexers.PythonLexer()
                if highlight_mode in {".py"}
                else pygments.lexers.XmlLexer()
                if highlight_mode in {".xml", ".svg", ".lbrn2"}
                else pygments.lexers.TexLexer()
            )

            html, style = highlight(content, lexer, formatter), self.css_content

        #        print(html)
        self.setHtml(
            f"""
<!doctype html>
<html lang="en">
  <meta charset="UTF-8"><head>{style}</head></meta>
  <body class="markdown-body markdown">{html}</body>
</html>
        """
        )

        return

        filename = Path(__file__).parent / "index.html"

        document = Document()
        download_manager = DownloadManager()

        channel = QWebChannel()
        channel.registerObject("content", document)

        # remote file
        markdown_url = QUrl.fromUserInput(
            "https://raw.githubusercontent.com/eyllanesc/stackoverflow/master/README.md"
        )
        # local file
        # markdown_url = QUrl.fromUserInput(/path/of/markdown.md)

        download_manager.finished.connect(document.set_text)
        download_manager.start_download(markdown_url)

        # view = QWebEngineView()
        self.page().setWebChannel(channel)
        url = QUrl.fromLocalFile(filename)
        view.load(url)



#!/usr/bin/env python3

"""QrM - Connect to reMarkable and modify contents
"""

# pylint: disable=invalid-name

import asyncio
import json
import logging
import os
import signal
import sys
from contextlib import suppress
from pathlib import Path

import qdarktheme
from PyQt6 import QtCore, QtGui, QtWidgets, uic
from PyQt6.QtWebEngineWidgets import QWebEngineView

from medit.utils import fs_changes, impatient, setup_logging, watchdog

CFG_FILE = "~/.medit.cfg"


def log() -> logging.Logger:
    """Returns the local logger"""
    return logging.getLogger("medit.ui")


def load_default(filename, default):
    with suppress(FileNotFoundError, json.JSONDecodeError):
        return json.load(open(os.path.expanduser(filename)))
    return default


# class FileFilterProxyModel(QtCore.QSortFilterProxyModel):
# def __init__(self, *args, **kwargs):
# super(FileFilterProxyModel, self).__init__(*args, **kwargs)
# self.excluded_files = []

# def setExcludedFiles(self, files):
# self.excluded_files = files
# self.invalidateFilter()

# def filterAcceptsRow(self, source_row, source_parent):
# source_model = self.sourceModel()
# index0 = source_model.index(source_row, 0, source_parent)
# filename = source_model.data(index0, QtCore.Qt.DisplayRole)

# return not filename in self.excluded_files


class MEditWindow(QtWidgets.QMainWindow):
    """The one and only application window"""

    @impatient
    def __init__(self, path: Path | None) -> None:
        super().__init__()
        uic.loadUi(Path(__file__).parent / "medit.ui", self)
        # self.setStyleSheet(styleSheet)
        # self.setStyleSheet("background-color: yellow;")

        # css =  open(Path(__file__).parent / "github-markdown-light.css").read()
        # print(css)
        # self.wv_rendered.setStyleSheet(css)
        # self.setAcceptDrops(True)
        # self.config = qrm_common.load_json(qrm_common.CFG_FILE)

        # https://stackoverflow.com/questions/66066115/render-markdown-with-pyqt5

        config = load_default(CFG_FILE, {})

        self.autosave_timer = QtCore.QTimer(self)
        self.autosave_timer.timeout.connect(self.on_autosave_timer_timeout)

        self.setGeometry(*config.get("window_geometry", (50, 50, 1000, 500)))
        self.base_dir = (
            path
            and path.exists()
            and (path if path.is_dir else path.parent)
            or Path(config.get("base_dir") or ".").absolute()
        )
        self.autoload_thread = QtCore.QThread(self)
        self.autoload_thread.run = lambda: self.async_stuff()
        self.autoload_thread.start()

        # self.fs_watcher = QtCore.QFileSystemWatcher([self.base_dir.as_posix()])
        # self.fs_watcher.fileChanged.connect(self.on_file_changed_externally)
        # self.fs_watcher.directoryChanged.connect(self.on_file_changed_externally)
        # self.fs_watcher.addPaths([self.base_dir.as_posix()])
        # print(self.fs_watcher.files())

        self.fs_model = QtGui.QFileSystemModel()
        self.fs_model.setRootPath(Path.home().as_posix())
        # self.fs_model.setNameFilters()
        self.fs_model.setNameFilters(["*.txt", "*.md"])
        self.open_file = None

        def extract_paths():
            if path:
                if not path.exists():
                    log().warning("%s does not exist", path)
                    return Path("."), None
                if path.is_dir():
                    return path, None
                return path.parent, path
            return Path(config.get("base_dir") or "."), config.get("open_file")

        self.base_dir, file_path = extract_paths()

        self.set_open_file(file_path)

        self.gb_splitter.setSizes(config.get("split_sizes", [1, 1, 1]))

        self.tv_files.setModel(self.fs_model)

        self.tv_files.selectionModel().selectionChanged.connect(self.on_tv_files_selectionChanged)

        self.tv_files.hideColumn(1)
        self.tv_files.hideColumn(2)
        self.tv_files.hideColumn(3)
        self.tv_files.setHeaderHidden(True)

        self.tv_files.setRootIndex(self.fs_model.index(self.base_dir.as_posix()))
        self.tv_files.expand(self.fs_model.index(self.base_dir.as_posix()))

        self.fs_model.sort(3, QtCore.Qt.SortOrder.DescendingOrder)

        if self.open_file:
            self.tv_files.setCurrentIndex(self.fs_model.index(self.open_file.absolute().as_posix()))
        else:
            self.tv_files.setCurrentIndex(self.fs_model.index(self.base_dir.absolute().as_posix()))

        self.show()

    def async_stuff(self):
        print("Async")
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        # terminator = threading.Event()

        # https://stackoverflow.com/a/71956673
        _tasks = {asyncio.ensure_future(self.watch_fs_changes())}
        loop.run_forever()

    @watchdog
    async def watch_fs_changes(self) -> None:
        """Observe changes to filesystem and mark walls dirty"""
        async for changed_path in fs_changes(self.base_dir, timeout=1, postpone=True):
            log().info("file %s changed", changed_path)
            QtCore.QMetaObject.invokeMethod(
                self,
                "on_file_changed_externally",
                QtCore.Qt.ConnectionType.QueuedConnection,
                QtCore.Q_ARG(str, changed_path.as_posix()),
            )

    def reset_timer(self):
        logging.debug("reset modification timer")
        self.autosave_timer.start(300)

    @impatient
    def render_content(self):
        self.wv_rendered.show(
            self.txt_editor.text(),
            self.open_file and self.open_file.suffix.strip("~"),
            self.base_dir,
        )

    @impatient
    def on_autosave_timer_timeout(self):
        """"""
        self.autosave_timer.stop()
        self.render_content()
        self.save()

    @impatient
    def save(self):
        # if not self.dirty:
        # return
        log().info("save to %s", self.open_file)
        # self.dirty = True
        text_to_save = self.txt_editor.text()
        if not text_to_save:
            log().warning("I don't dare to overwrite with empty content..")
            return
        open(self.open_file, "w").write(text_to_save)

    @impatient
    def set_open_file(self, path):
        self.open_file = None
        block_state = self.txt_editor.blockSignals(True)
        selected_file = Path(path) if path else None
        if not path:
            return
        try:
            self.txt_editor.openFile(selected_file)
            self.open_file = selected_file
        except UnicodeDecodeError:
            self.txt_editor.setText("")
        finally:
            self.txt_editor.blockSignals(block_state)
            self.setWindowTitle(str(self.open_file))
            QtWidgets.QApplication.instance().setApplicationName(
                f"medit - {self.open_file and self.open_file.name or ''}"
            )

        self.render_content()

    @impatient
    def on_tv_files_selectionChanged(self, selection1, selection2) -> None:
        # print(Path(self.fs_model.filePath(selection.indexes()[0])))
        try:
            selected_path = Path(self.fs_model.filePath(selection1.indexes()[0]))
            if selected_path.is_file():
                self.set_open_file(selected_path)
        except IndexError:
            print("FILE REMOVED?")

    def on_txt_editor_textChanged(self) -> None:
        self.reset_timer()

    @impatient
    @QtCore.pyqtSlot(str)
    def on_file_changed_externally(self, file) -> None:
        print(file)

    @impatient
    def event(self, event: QtCore.QEvent) -> bool:
        # if event.type() == QtCore.QEvent.DragEnter:
        # if any(
        # Path(u.url()).suffix.lower() in {".pdf", ".epub"} for u in event.mimeData().urls()
        # ):
        # event.accept()
        # elif event.type() == QtCore.QEvent.Drop:
        # urls = [
        # path
        # for u in event.mimeData().urls()
        # if (path := Path(u.url())).suffix.lower() in {".pdf", ".epub"}
        # ]
        # print(urls)

        # elif not event.type() in {
        # QtCore.QEvent.UpdateRequest,
        # QtCore.QEvent.Paint,
        # QtCore.QEvent.Enter,
        # QtCore.QEvent.HoverEnter,
        # QtCore.QEvent.HoverMove,
        # QtCore.QEvent.HoverLeave,
        # QtCore.QEvent.KeyPress,
        # QtCore.QEvent.KeyRelease,
        # QtCore.QEvent.DragMove,
        # QtCore.QEvent.DragLeave,
        # }:
        ## log().warn("unknown event: %r %r", event.type(), event)
        # pass
        return super().event(event)

    @impatient
    def closeEvent(self, _event: QtGui.QCloseEvent) -> None:
        """save state before shutting down"""
        logging.info("got some closish signal, bye")
        # self.autoload_thread.terminate()
        # self.autoload_thread.wait()
        # self.save()
        g = self.geometry()
        # splitter =
        json.dump(
            {
                "window_geometry": (g.x(), g.y(), g.width(), g.height()),
                "editor_view_state": self.txt_editor.view_state(),
                "open_file": self.open_file and self.open_file.as_posix(),
                "base_dir": self.base_dir and self.base_dir.as_posix(),
                "split_sizes": self.gb_splitter.sizes(),
            },
            open(os.path.expanduser(CFG_FILE), "w"),
        )


def main(path: Path) -> None:
    """Typical PyQt5 boilerplate main entry point"""

    setup_logging()
    app = QtWidgets.QApplication(sys.argv)

    qdarktheme.setup_theme("light")

    window = MEditWindow(path)

    for s in (signal.SIGABRT, signal.SIGINT, signal.SIGSEGV, signal.SIGTERM):
        signal.signal(s, lambda signal, frame: window.close())

    # catch the interpreter every now and then to be able to catch signals
    timer = QtCore.QTimer()
    timer.start(200)
    timer.timeout.connect(lambda: None)

    sys.exit(app.exec())


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""Markup Editor
"""

import importlib.metadata
import logging
import shutil
import sys
from argparse import ArgumentParser
from argparse import Namespace as Args
from collections.abc import Sequence
from contextlib import suppress
from pathlib import Path

from medit.ui import main as ui_main
from medit.utils import setup_logging


def parse_args(argv: Sequence[str] | None = None) -> Args:
    """Cool git like multi command argument parser"""
    parser = ArgumentParser(__doc__)
    parser.add_argument("--verbose", "-v", action="store_true")
    parser.add_argument("--info", action="store_true")

    parser.set_defaults(func=fn_ui)
    parser.add_argument("path", type=Path, nargs="?")

    return parser.parse_args(argv)


def logger() -> logging.Logger:
    """Named logger"""
    return logging.getLogger("medit.cli")


def extract_version() -> str:
    """Returns either the version of installed package or the one
    found in nearby pyproject.toml"""
    with suppress(FileNotFoundError, StopIteration):
        with open(
            Path(__file__).parent.parent / "pyproject.toml", encoding="utf-8"
        ) as pyproject_toml:
            version = (
                next(line for line in pyproject_toml if line.startswith("version"))
                .split("=")[1]
                .strip("'\"\n ")
            )
            return f"{version}-dev"
    return importlib.metadata.version(__name__.split(".", maxsplit=1)[0])


__version__ = extract_version()


def shorten_home(path: Path | str) -> Path:
    """Reverse of expanduser"""
    return Path(Path(path).as_posix().replace(str(Path.home()), "~"))


def fn_info(_args: Args) -> None:
    """Entry point `info`"""
    print(f"Version: {__version__} (at {shorten_home(Path(__file__).parent)})")
    print(
        f"Python: {'.'.join(map(str, sys.version_info[:3]))}"
        f" (at {shorten_home(sys.executable)})"
    )


def fn_ui(args: Args) -> None:
    ui_main(args.path if hasattr(args, "path") else None)


def main() -> int:
    """Entry point for everything else"""
    args = parse_args()
    return fn_info(args) if args.info else fn_ui(args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# https://stackoverflow.com/questions/40002373/qscintilla-based-text-editor-in-pyqt5-with-clickable-functions-and-variables

from pathlib import Path

from PyQt6 import Qsci, QtCore, QtGui
from PyQt6.QtGui import QImage, QPainter


class MEditor(Qsci.QsciScintilla):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setIndentationsUseTabs(False)
        self.setIndentationWidth(4)

        font = QtGui.QFont()
        font.setFamily("Source Code Pro Semibold")
        if font.exactMatch():
            print("Font set successfully")
        else:
            print("Font not found, using closest match", font.family())
        font.setFixedPitch(True)
        font.setPointSize(20)
        self.setFont(font)
        self.setMarginsFont(font)
        # self.zoomIn()
        # self.zoomOut()

        # Margin 0 is used for line numbers
        fontmetrics = QtGui.QFontMetrics(font)
        self.setMarginsFont(font)
        self.setMarginWidth(0, fontmetrics.boundingRect("000").width() + 6)
        # self.setMarginLineNumbers(0, True)
        self.setMarginsBackgroundColor(QtGui.QColor("#eee8d5"))
        self.setMarginsForegroundColor(QtGui.QColor("#93a1a1"))
        # self.setMarginBackgroundColor(1, QtGui.QColor("lightgray"))

        # self._marker = None
        # Clickable margin 1 for showing markers
        # self.setMarginSensitivity(1, True)
        # self.connect(self,
        #    SIGNAL('marginClicked(int, int, Qt::KeyboardModifiers)'),
        #    self.on_margin_clicked)
        # self.markerDefine(QsciScintilla.RightArrow, self.ARROW_MARKER_NUM)
        # self.setMarkerBackgroundColor(QColor("#ee1111"), self.ARROW_MARKER_NUM)

        self.setBraceMatching(self.BraceMatch.SloppyBraceMatch)

        # Current line visible with special background color
        self.setCaretLineVisible(True)
        self.setCaretLineBackgroundColor(QtGui.QColor("#eee8d5"))

        self.SendScintilla(Qsci.QsciScintilla.SCI_STYLESETFONT, 1, "Courier".encode())

        # Don't want to see the horizontal scrollbar at all
        # Use raw message to Scintilla here (all messages are documented
        # here: http://www.scintilla.org/ScintillaDoc.html)
        self.SendScintilla(Qsci.QsciScintilla.SCI_SETHSCROLLBAR, 0)

        # self.setWrapMode(self.WrapMode.WrapWord)
        self.setWrapMode(self.WrapMode.WrapNone)
        self.setEolMode(self.EolMode.EolUnix)
        self.setEolVisibility(False)
        self.setWhitespaceVisibility(self.WhitespaceVisibility.WsVisible)

        # self.image = QImage("logo_mk.png")
        # painter = QPainter()
        # painter.begin(self.image)
        # painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_DestinationIn)
        # painter.fillRect(self.image.rect(), QtGui.QColor(0, 0, 0, 100))
        # painter.end()

    # def paintEvent(self, e):
    # painter = QPainter()
    # painter.begin(self.viewport())
    # painter.drawImage(QtCore.QPoint(0, 0), self.image)
    # painter.end()
    # super().paintEvent(e)

    def openFile(self, path):
        suffix = path.suffix.lower().strip("~")
        lexer = (
            Qsci.QsciLexerMarkdown()
            if suffix in {".md", ".markdown"}
            else Qsci.QsciLexerYAML()
            if suffix in {".yaml", ".yml"}
            else Qsci.QsciLexerBash()
            if suffix in {".sh"}
            else Qsci.QsciLexerPython()
            if suffix in {".py"}
            else Qsci.QsciLexerXML()
            if suffix in {".xml", ".svg"}
            else None
        )
        if lexer:
            lexer.setFont(self.font())
            # lexer.setPaper(QtGui.QColor("#073642"))
            # lexer.setDefaultPaper(QtGui.QColor("#073642"))
            lexer.setPaper(QtGui.QColor("#fdf6e3"))
            lexer.setDefaultPaper(QtGui.QColor("#fdf6e3"))

        self.setLexer(lexer)

        with open(path) as textFile:
            self.setText(textFile.read())

    def view_state(self):
        # In Qt5/PyQt5 how do I restore exact visible area and cursor
        #  position of QTextEdit?:
        # https://stackoverflow.com/questions/67751888

        # relative approach
        # try:
        # hPos = hBar.value() / hBar.maximum()
        # except ZeroDivisionError:
        # hPos = 0
        # try:
        # vPos = vBar.value() / vBar.maximum()
        # except ZeroDivisionError:
        # vPos = 0

        state = {
            "cursor_position": self.getCursorPosition(),
            "zoom": 0,  # TODO: how to get?
            "hbar_pos": self.horizontalScrollBar().value(),
            "vbar_pos": self.verticalScrollBar().value(),
        }
        return state

    def set_view_state(self, state):
        if not state:
            return
        self.setCursorPosition(*state["cursor_position"])
        # hBar.setValue(hPos * hBar.maximum())  # relative approach
        # vBar.setValue(vPos * vBar.maximum())
        self.horizontalScrollBar().setValue(state["hbar_pos"])
        self.verticalScrollBar().setValue(state["vbar_pos"])
